@@ -5,7 +5,6 @@ import matplotlib.pyplot as plt
 from pynvml import *
 import numpy as np
 
-
 class StatsRecorder(object):
     def __init__(self):
         self.stopped = False
@@ -14,6 +13,9 @@ class StatsRecorder(object):
     def start(self, interval=1):
         Thread(target=self._update, args=([interval])).start()
         return self
+    
+    def _run_command(self, cmd):
+        return subprocess.getoutput(cmd)
 
     def _update(self, interval=1):
         t = 0
@@ -42,6 +44,8 @@ class NVStatsRecorder(StatsRecorder):
     Used to record generic NVIDIA GPU metrics over a time period.
     Args:
         gpu_index (int): Index of GPU to instrument (starts from 0)
+        tensor_util (bool): Collect Tensor Core utilization
+        sudo_password (str): Superuser password required to collect Tensor Core utilization
     Usage:
         ```python
         nv_stats_recorder = NVStatsRecorder(gpu_index=0)
@@ -51,10 +55,23 @@ class NVStatsRecorder(StatsRecorder):
         gpu_data = nv_stats_recorder.get_data()
         ```
     """
-    def __init__(self, gpu_index=0):
+    def __init__(self, gpu_index=0, tensor_util=False, sudo_password=""):
         self.stopped = False
         self.time_history = []
         self.sm_util_history = []
+        self.tensor_util = tensor_util
+        if self.tensor_util:
+            print("\n[  IMPORTANT NOTICE                                                 ]")
+            print("`tensor_util` is set to `True`")
+            print("This enables Tensor Core utilization metrics collected via DCGM 1.7+")
+            print("Supported GPUs are V100, T4 on Tesla-ready driver (418, 440+) only!")
+            self.sudo_password = str(sudo_password)
+            if len(self.sudo_password) < 2:
+                print("This requires starting DCGM with superuser permissions! Please specify `sudo_password` argument.")
+            output = self._start_dcgm_profiler()
+            print("Attempted to start DCGM:", output.strip())
+            print("[   END NOTICE                                                      ]\n")
+        self.tensor_util_history = []
         self.sm_clocks_history = []
         self.mem_clocks_history = []
         self.mem_util_history = []
@@ -65,6 +82,7 @@ class NVStatsRecorder(StatsRecorder):
         self.throttle_reasons = []
         nvmlInit()
         self.handle = nvmlDeviceGetHandleByIndex(gpu_index)
+        self.gpu_index = str(gpu_index)
         # calculate max PCIE bandwidth
         try:
             pcie_gen = nvmlDeviceGetMaxPcieLinkGeneration(self.handle)
@@ -95,8 +113,27 @@ class NVStatsRecorder(StatsRecorder):
             "max_mem_clock": nvmlDeviceGetMaxClockInfo(self.handle, type=2),
             "max_pcie_bandwidth": int(pcie_bandwidth),
             "max_power": int(nvmlDeviceGetEnforcedPowerLimit(self.handle)),
+            "temp_max": int(nvmlDeviceGetTemperatureThreshold(self.handle, 0)),
+            "temp_slow": int(nvmlDeviceGetTemperatureThreshold(self.handle, 1)),
         }
 
+    def _start_dcgm_profiler(self):
+        return self._run_command(
+            "echo '" + self.sudo_password + "' | sudo -S nv-hostengine"
+        )
+    
+    def _get_tensor_pipe_active(self):
+        try:
+            output = self._run_command(
+                "dcgmi dmon -e 1004 -c 1 -i " + self.gpu_index
+            )
+            output = float(output.split(" "+self.gpu_index+"  ")[1].strip()) * 100
+            return output
+        except Exception as e:
+            print(e)
+            print(output)
+            return 0.0
+        
     def _get_throttle_reason(self, bitmask):
         bitmask = str(int(bitmask))
         return self.throttle_reason_map[bitmask]
@@ -117,6 +154,7 @@ class NVStatsRecorder(StatsRecorder):
                 nvmlShutdown()
                 return
             else:
+                st = time.time()
                 util = nvmlDeviceGetUtilizationRates(self.handle)
                 self.sm_util_history.append(util.gpu)
                 self.mem_util_history.append(util.memory)
@@ -153,11 +191,16 @@ class NVStatsRecorder(StatsRecorder):
                 ):
                     print("Detected GPU throttle:", throttle_reason_str)
                     self.throttle_reasons.append((throttle_reason_str, t))
-                time.sleep(interval)
-            self.time_history.append(t)
-            t += interval
+                if self.tensor_util:
+                    self.tensor_util_history.append(self._get_tensor_pipe_active())
+                et = time.time()
+                while (et-st) < interval:
+                    time.sleep(interval/5)
+                    et = time.time()
+                t += (et-st)
+                self.time_history.append(t)
 
-    def plot_gpu_util(self, smooth=2, figsize=(10,8), dpi=100, outpath=None):
+    def plot_gpu_util(self, smooth=2, figsize=(10,6), dpi=100, outpath=None):
         gpu_data = self.get_data()
         trunc = smooth - 1
         t_len = len(gpu_data["time_history"])
@@ -175,6 +218,11 @@ class NVStatsRecorder(StatsRecorder):
         plt.clf()
         plt.figure(figsize=figsize, dpi=dpi)
         plt.plot(time_history, sm_util_history, label="SM Util", color="g")
+        if self.tensor_util:
+            tensor_util_history = self._moving_average(gpu_data["tensor_util_history"], smooth)[
+                : t_len - trunc
+            ]
+            plt.plot(time_history, tensor_util_history, label="TC Util", color="#76b900")
         plt.plot(time_history, mem_util_history, label="Mem I/O", color="c")
         plt.plot(time_history, gpu_data["mem_occupy_history"][trunc:t_len], label="Mem Usage", color="y")
         plt.plot(time_history, pwr_history_history, label="Power", color="m")
@@ -187,6 +235,7 @@ class NVStatsRecorder(StatsRecorder):
                 listed.append(label)
             else:
                 plt.axvline(item[1], color="r", linestyle="--", linewidth=1)
+        plt.axhline(0, color="k", linewidth=1)
         plt.title("GPU Utilization")
         plt.ylabel("%")
         plt.xlabel("Time")
@@ -194,6 +243,36 @@ class NVStatsRecorder(StatsRecorder):
         if outpath:
             plt.savefig(outpath)
         plt.show()
+        
+    def plot_gpu_temp(self, smooth=2, figsize=(10,6), dpi=100, outpath=None):
+        gpu_data = self.get_data()
+        trunc = smooth - 1
+        t_len = len(gpu_data["time_history"])
+        time_history = gpu_data["time_history"][trunc:t_len]
+        temp_history = self._moving_average(gpu_data["temp_history"], smooth)[
+            : t_len - trunc
+        ]
+        plt.clf()
+        plt.figure(figsize=figsize, dpi=dpi)
+        plt.plot(time_history, temp_history, label="Temperature", color="b")
+        listed = []
+        for item in gpu_data["throttle_reasons"]:
+            label = str(item[0])
+            if label not in listed:
+                plt.axvline(item[1], color="r", linestyle="--", linewidth=1, label=label)
+                listed.append(label)
+            else:
+                plt.axvline(item[1], color="r", linestyle="--", linewidth=1)
+        plt.axhline(self.device_data["temp_slow"], color="y", linestyle="--", label="Slowdown Temp")
+        plt.axhline(self.device_data["temp_max"], color="r", linestyle="--", label="Max Temp")
+        plt.axhline(0, color="k", linewidth=1)
+        plt.title("GPU Temperature")
+        plt.ylabel("degree C")
+        plt.xlabel("Time")
+        plt.legend(loc="upper left")
+        if outpath:
+            plt.savefig(outpath)
+        plt.show()
 
     def get_data(self):
         t_len = len(self.time_history)
@@ -201,6 +280,7 @@ class NVStatsRecorder(StatsRecorder):
             "device_data": self.device_data,
             "time_history": self.time_history[:t_len],
             "sm_util_history": self.sm_util_history[:t_len],
+            "tensor_util_history": self.tensor_util_history[:t_len],
             "sm_clocks_history": self.sm_clocks_history[:t_len],
             "mem_util_history": self.mem_util_history[:t_len],
             "mem_clocks_history": self.mem_clocks_history[:t_len],
@@ -211,6 +291,31 @@ class NVStatsRecorder(StatsRecorder):
             "throttle_reasons": self.throttle_reasons,
         }
         return data
+    
+    def summary_seq(self, data_seq, idle_threshold=0.25):
+        data_min = min(data_seq)
+        data_max = max(data_seq)
+        data_nonidle = []
+        for d in data_seq:
+            if d > idle_threshold:
+                data_nonidle.append(d)
+        steady_state = sum(data_nonidle)/len(data_nonidle)
+        return (round(data_min, 1), round(steady_state, 1), round(data_max, 1))
+    
+    def summary(self):
+        print("Summary (%)")
+        print("===========")
+        print("Metric  | Min    | Max    | Non-idle Avg")
+        gpu_data = self.get_data()
+        sm_summary = self.summary_seq(gpu_data["sm_util_history"])
+        print("SM Util |", sm_summary[0], "|", sm_summary[2], "|", sm_summary[1])
+        mem_summary = self.summary_seq(gpu_data["mem_util_history"])
+        print("Mem I/O |", mem_summary[0], "|", mem_summary[2], "|", mem_summary[1])
+        pwr_summary = self.summary_seq(gpu_data["pwr_history"])
+        print("Power   |", pwr_summary[0], "|", pwr_summary[2], "|", pwr_summary[1])
+        if self.tensor_util:
+            tensor_summary = self.summary_seq(gpu_data["tensor_util_history"])
+            print("Tensor  |", tensor_summary[0], "|", tensor_summary[2], "|", tensor_summary[1])
 
 
 class NVLinkStatsRecorder(StatsRecorder):
@@ -222,9 +327,6 @@ class NVLinkStatsRecorder(StatsRecorder):
         self.sudo_password = str(sudo_password)
         self.colors = ["b", "g", "c", "m", "y", "b", "g", "c", "m", "y", "b", "g", "c", "m", "y", "b", "g", "c", "m", "y"]
 
-    def _run_command(self, cmd):
-        return subprocess.getoutput(cmd)
-
     def _start_nvlink_counter(self):
         return self._run_command(
             "echo '" + self.sudo_password + "' | sudo -S nvidia-smi nvlink -sc 0bz"
@@ -240,7 +342,7 @@ class NVLinkStatsRecorder(StatsRecorder):
             "echo '" + self.sudo_password + "' | sudo -S nvidia-smi nvlink -g 0"
         )
 
-    def _parse_nvlink_output(self, output):
+    def _parse_nvlink_output(self, output, interval):
         """
         Returns the NVLink throughput across all GPUs
         """
@@ -260,10 +362,10 @@ class NVLinkStatsRecorder(StatsRecorder):
                             rx = int(kbyte.split(": ")[-1])
                             list_rx.append(rx)
                 total_bw = sum(list_tx) + sum(list_rx)
-                gpu_bw[str(i)] = total_bw / 1e6
+                gpu_bw[str(i)] = total_bw / 1e6 / interval
         return gpu_bw
 
-    def plot_nvlink_traffic(self, smooth=2, figsize=(10,8), dpi=100, outpath=None):
+    def plot_nvlink_traffic(self, smooth=2, figsize=(10,6), dpi=100, outpath=None):
         nvlink_data = self.get_data()
         trunc = smooth - 1
         t_len = len(nvlink_data["time_history"])
@@ -291,7 +393,7 @@ class NVLinkStatsRecorder(StatsRecorder):
                 plt.fill_between(time_history, per_gpu_data+cum_gpu_data, cum_gpu_data, color=color, alpha=0.5)
                 cum_gpu_data += per_gpu_data
                 plt.plot(time_history, cum_gpu_data, color=color, label="GPU"+str(gpu))
-        plt.axhline(0, color="k")
+        plt.axhline(0, color="k", linewidth=1)
         plt.legend(loc="upper left")
         plt.title("NVLink Traffic")
         plt.ylabel("GB/s")
@@ -301,24 +403,27 @@ class NVLinkStatsRecorder(StatsRecorder):
         plt.show()
 
     def start(self, interval=2):
-        self._start_nvlink_counter()
-        self._reset_nvlink_counter()
         Thread(target=self._update, args=([interval])).start()
         return self
 
     def _update(self, interval):
         t = 0
+        counter = {}
+        for i in self.gpus:
+            counter[str(i)] = 0
         while True:
             if self.stopped:
                 return
             else:
-                output = self._read_nvlink_counter()
+                st = time.time()
                 self._reset_nvlink_counter()
-                counter = self._parse_nvlink_output(output)
                 self.nvlink_history.append(counter)
                 time.sleep(interval)
-            self.time_history.append(t)
-            t += interval
+                output = self._read_nvlink_counter()
+                counter = self._parse_nvlink_output(output, interval)
+                et = time.time()
+                t += (et-st)
+                self.time_history.append(t)
 
     def get_data(self):
         data = {
